@@ -408,6 +408,8 @@ Shows success or final failure with raw error."
   "Handle setStatus method from EVENT."
   (let ((key (plist-get event :statusKey))
         (text (plist-get event :statusText)))
+    (when text
+      (setq text (ansi-color-filter-apply text)))
     (if text
         (setq pi-coding-agent--extension-status
               (cons (cons key text)
@@ -462,7 +464,7 @@ Shown when the session starts without a configured model/API key."
 
 (defun pi-coding-agent--cleanup-on-kill ()
   "Clean up resources when chat buffer is killed.
-Also kills the linked input buffer.
+Also kills the linked input buffer and fontification cache buffers.
 
 Note: This runs from `kill-buffer-hook', which executes AFTER the kill
 decision is made.  For proper cancellation support, use `pi-coding-agent-quit'
@@ -472,6 +474,7 @@ which asks upfront before any buffers are touched."
       (pi-coding-agent--unregister-display-handler pi-coding-agent--process)
       (when (process-live-p pi-coding-agent--process)
         (delete-process pi-coding-agent--process)))
+    (pi-coding-agent--kill-fontify-buffers)
     (when (and pi-coding-agent--input-buffer (buffer-live-p pi-coding-agent--input-buffer))
       (let ((input-buf pi-coding-agent--input-buffer))
         (pi-coding-agent--set-input-buffer nil) ; break cycle before kill
@@ -850,7 +853,8 @@ Uses `font-lock-face' to survive gfm-mode refontification."
            name))))))
 
 (defun pi-coding-agent--display-tool-start (tool-name args)
-  "Display header for tool TOOL-NAME with ARGS and create overlay."
+  "Insert tool header for TOOL-NAME with ARGS and create pending overlay.
+Records the header-end position for later content insertion."
   (let* ((header-display (pi-coding-agent--tool-header tool-name args))
          (path (pi-coding-agent--tool-path args))
          (inhibit-read-only t))
@@ -975,10 +979,13 @@ when multi-line constructs (docstrings, block comments) start
 above the visible window.
 
 Inhibits modification hooks to prevent jit-lock from scanning the
-full buffer on each delta."
+full buffer on each delta.  Skips the delete+reinsert cycle when the
+visible tail content is unchanged from the previous delta."
   (when (and pi-coding-agent--pending-tool-overlay
              raw-text
              (not (string-empty-p raw-text)))
+    ;; Always sync fontify buffer (incremental, cheap) regardless of
+    ;; whether the display will update — the buffer must stay current.
     (when lang
       (pi-coding-agent--fontify-sync raw-text lang))
     (let* ((tail-result
@@ -1000,38 +1007,52 @@ full buffer on each delta."
             (or has-hidden
                 (and truncation
                      (> (plist-get truncation :hidden-lines) 0))))
-           (inhibit-read-only t)
-           (inhibit-modification-hooks t))
-      ;; Nothing to display yet (e.g., first delta before any newline)
-      (unless (string-empty-p display-content)
-        (pi-coding-agent--with-scroll-preservation
-          (save-excursion
-            (let* ((ov-end (overlay-end pi-coding-agent--pending-tool-overlay))
-                   (header-end (overlay-get pi-coding-agent--pending-tool-overlay
-                                            'pi-coding-agent-header-end)))
-              ;; Delete previous streaming content (everything after header)
-              (when (and header-end (< header-end ov-end))
-                (delete-region header-end ov-end))
-              ;; Insert new streaming content
-              (goto-char (overlay-end pi-coding-agent--pending-tool-overlay))
-              (when show-hidden-indicator
-                (insert (propertize "... (earlier output)\n"
-                                    'face 'pi-coding-agent-collapsed-indicator)))
-              (let ((content-start (point)))
-                (insert (if lang
-                            display-content
-                          (pi-coding-agent--render-tool-content
-                           display-content nil)))
-                (insert "\n")
-                ;; Mark pre-fontified content so jit-lock won't override
-                ;; our syntax faces with gfm-mode faces on redisplay.
-                ;; Also layer markdown-code-face underneath so the text
-                ;; uses fixed-pitch font, matching completed code blocks.
-                (when lang
-                  (put-text-property content-start (point)
-                                     'fontified t)
-                  (add-face-text-property content-start (point)
-                                          'markdown-code-face t))))))))))
+           ;; Compare plain text to cached value — skip if unchanged.
+           ;; Property-stripped comparison is correct: the text determines
+           ;; whether the preview changed.  Font-lock properties may shift
+           ;; cosmetically but that's not worth a full redraw.
+           (display-text (substring-no-properties display-content))
+           (cache-key (if show-hidden-indicator
+                         (concat "H:" display-text)
+                       display-text))
+           (last-tail (overlay-get pi-coding-agent--pending-tool-overlay
+                                   'pi-coding-agent-last-tail)))
+      ;; Skip display when the visible tail is unchanged
+      (unless (or (string-empty-p display-content)
+                  (equal cache-key last-tail))
+        (let ((inhibit-read-only t)
+              (inhibit-modification-hooks t))
+          (pi-coding-agent--with-scroll-preservation
+            (save-excursion
+              (let* ((ov-end (overlay-end pi-coding-agent--pending-tool-overlay))
+                     (header-end (overlay-get pi-coding-agent--pending-tool-overlay
+                                              'pi-coding-agent-header-end)))
+                ;; Delete previous streaming content (everything after header)
+                (when (and header-end (< header-end ov-end))
+                  (delete-region header-end ov-end))
+                ;; Insert new streaming content
+                (goto-char (overlay-end pi-coding-agent--pending-tool-overlay))
+                (when show-hidden-indicator
+                  (insert (propertize "... (earlier output)\n"
+                                      'face 'pi-coding-agent-collapsed-indicator)))
+                (let ((content-start (point)))
+                  (insert (if lang
+                              display-content
+                            (pi-coding-agent--render-tool-content
+                             display-content nil)))
+                  (insert "\n")
+                  ;; Mark pre-fontified content so jit-lock won't override
+                  ;; our syntax faces with gfm-mode faces on redisplay.
+                  ;; Also layer markdown-code-face underneath so the text
+                  ;; uses fixed-pitch font, matching completed code blocks.
+                  (when lang
+                    (put-text-property content-start (point)
+                                       'fontified t)
+                    (add-face-text-property content-start (point)
+                                            'markdown-code-face t))))))
+          ;; Cache the displayed content for next comparison
+          (overlay-put pi-coding-agent--pending-tool-overlay
+                       'pi-coding-agent-last-tail cache-key))))))
 
 (defun pi-coding-agent--display-tool-update (partial-result)
   "Display PARTIAL-RESULT as streaming output in pending tool overlay.
@@ -1057,9 +1078,28 @@ Returns the rendered string."
       (pi-coding-agent--wrap-in-src-block content lang)
     (propertize content 'face 'pi-coding-agent-tool-output)))
 
-(defsubst pi-coding-agent--fontify-buffer-name (lang)
-  "Return the name of the fontification cache buffer for LANG."
-  (format " *pi-fontify:%s*" lang))
+(defun pi-coding-agent--fontify-get-buffer (lang)
+  "Return the fontification cache buffer for LANG in the current session.
+Looks up the buffer-local `pi-coding-agent--fontify-buffers' hash table.
+Returns nil if no buffer exists for LANG.  Removes stale entries
+for buffers that have been killed externally."
+  (when pi-coding-agent--fontify-buffers
+    (let ((buf (gethash lang pi-coding-agent--fontify-buffers)))
+      (cond
+       ((null buf) nil)
+       ((buffer-live-p buf) buf)
+       (t (remhash lang pi-coding-agent--fontify-buffers) nil)))))
+
+(defun pi-coding-agent--fontify-get-or-create-buffer (lang)
+  "Return or create the fontification cache buffer for LANG.
+Uses the buffer-local hash table to track per-session buffers.
+Returns nil if called outside a chat buffer (no hash table)."
+  (when pi-coding-agent--fontify-buffers
+    (or (pi-coding-agent--fontify-get-buffer lang)
+        (let ((buf (generate-new-buffer
+                    (format " *pi-fontify:%s:%s*" lang (buffer-name)))))
+          (puthash lang buf pi-coding-agent--fontify-buffers)
+          buf))))
 
 (defun pi-coding-agent--fontify-reset (args)
   "Clear the fontification buffer for the language implied by ARGS.
@@ -1068,7 +1108,7 @@ buffer, preventing stale content from a previous call from being
 treated as a matching prefix during incremental sync."
   (when-let* ((lang (pi-coding-agent--path-to-language
                      (pi-coding-agent--tool-path args)))
-              (buf (get-buffer (pi-coding-agent--fontify-buffer-name lang))))
+              (buf (pi-coding-agent--fontify-get-buffer lang)))
     (with-current-buffer buf
       (erase-buffer))))
 
@@ -1081,9 +1121,8 @@ incorrect keyword matching on partial tokens.
 The buffer always accumulates content regardless of whether the
 language mode is available, so `pi-coding-agent--fontify-buffer-tail'
 can extract the tail even for languages without an installed mode."
-  (condition-case nil
-      (let ((buf (get-buffer-create
-                  (pi-coding-agent--fontify-buffer-name lang))))
+  (condition-case err
+      (when-let* ((buf (pi-coding-agent--fontify-get-or-create-buffer lang)))
         (with-current-buffer buf
           ;; Activate language mode once (best-effort for fontification).
           (let ((mode (and lang (markdown-get-lang-mode lang))))
@@ -1117,7 +1156,8 @@ can extract the tail even for languages without an installed mode."
               (ignore-errors
                 (font-lock-default-fontify-region
                  (point-min) (point-max) nil)))))))
-    (error nil)))
+    (error
+     (message "pi-coding-agent: fontify-sync error for %s: %S" lang err))))
 
 (defun pi-coding-agent--fontify-buffer-tail (lang n)
   "Extract last N non-blank complete lines from the LANG fontification buffer.
@@ -1135,7 +1175,7 @@ Returns (CONTENT . HAS-HIDDEN) where CONTENT is a string with text
 properties preserved.  HAS-HIDDEN is non-nil when earlier lines
 exist above the returned tail.
 Returns nil if the buffer doesn't exist or has no complete lines."
-  (let ((buf (get-buffer (pi-coding-agent--fontify-buffer-name lang))))
+  (let ((buf (pi-coding-agent--fontify-get-buffer lang)))
     (when (and buf (> (buffer-size buf) 0))
       (with-current-buffer buf
         ;; Find end of last complete line (just before the trailing
@@ -1162,6 +1202,17 @@ Returns nil if the buffer doesn't exist or has no complete lines."
               (cons (mapconcat #'identity line-strings "\n")
                     (> (point) (point-min))))))))))
 
+(defun pi-coding-agent--kill-fontify-buffers ()
+  "Kill all fontification cache buffers for the current session.
+Iterates the buffer-local `pi-coding-agent--fontify-buffers' hash table
+and kills each buffer, then clears the table."
+  (when pi-coding-agent--fontify-buffers
+    (maphash (lambda (_lang buf)
+               (when (buffer-live-p buf)
+                 (kill-buffer buf)))
+             pi-coding-agent--fontify-buffers)
+    (clrhash pi-coding-agent--fontify-buffers)))
+
 (defun pi-coding-agent--display-tool-end (tool-name args content details is-error)
   "Display result for TOOL-NAME and update overlay face.
 ARGS contains tool arguments, CONTENT is a list of content blocks.
@@ -1661,14 +1712,57 @@ correct context, not blocks from earlier in the buffer."
         (funcall orig-fun prop (if lim (max lim limit) limit)))
     (funcall orig-fun prop lim)))
 
+(defun pi-coding-agent--restore-tool-properties (beg end)
+  "Strip markdown text properties from the pending tool overlay in BEG..END.
+Removes properties that gfm-mode fontification applies to markup
+patterns in tool output:
+- `display' (\"\"): hides # in headings
+- `invisible' (markdown-markup): hides ** __ and heading markup
+- `font-lock-multiline': causes fontification region extensions
+- `face': overrides tool faces with markdown heading/bold faces
+Restores intended faces for both the header and content regions."
+  (when-let* ((ov pi-coding-agent--pending-tool-overlay)
+              (ov-start (overlay-start ov))
+              (ov-end (overlay-end ov))
+              (header-end-marker (overlay-get ov 'pi-coding-agent-header-end))
+              (header-end (marker-position header-end-marker)))
+    (when (and (< beg ov-end) (> end ov-start))
+      (let ((inhibit-read-only t))
+        ;; Header: restore face from font-lock-face (varies per span)
+        (let ((hdr-beg (max beg ov-start))
+              (hdr-end (min end header-end)))
+          (when (< hdr-beg hdr-end)
+            (remove-text-properties
+             hdr-beg hdr-end
+             '(display nil invisible nil font-lock-multiline nil))
+            (let ((pos hdr-beg))
+              (while (< pos hdr-end)
+                (let* ((fl-face (get-text-property pos 'font-lock-face))
+                       (next (or (next-single-property-change
+                                  pos 'font-lock-face nil hdr-end)
+                                 hdr-end)))
+                  (when fl-face
+                    (put-text-property pos next 'face fl-face))
+                  (setq pos next))))
+            (put-text-property hdr-beg hdr-end 'fontified t)))
+        ;; Content: uniform tool-output face
+        (let ((cnt-beg (max beg header-end))
+              (cnt-end (min end ov-end)))
+          (when (< cnt-beg cnt-end)
+            (remove-text-properties
+             cnt-beg cnt-end
+             '(display nil invisible nil font-lock-multiline nil))
+            (put-text-property cnt-beg cnt-end 'face
+                               'pi-coding-agent-tool-output)
+            (put-text-property cnt-beg cnt-end 'fontified t)))))))
+
 (defun pi-coding-agent--fontify-streaming-region ()
   "Fontify newly streamed message text incrementally.
 Called by idle timer during streaming.  Only fontifies message text
 that hasn't been fontified yet, tracked via the variable
 `pi-coding-agent--last-fontified-pos'.  Skips the pending tool
-overlay region — tool content is pre-fontified in a temp buffer
-and marked with the `fontified' property to prevent jit-lock from
-overriding it with gfm-mode faces."
+overlay region to avoid applying gfm-mode faces to tool content
+via `font-lock-ensure' (which is not cleaned up by jit-lock)."
   (when (and pi-coding-agent--message-start-marker
              pi-coding-agent--streaming-marker
              (marker-position pi-coding-agent--message-start-marker)
